"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.bootstrapCore = void 0;
const redux_clean_architecture_1 = require("redux-clean-architecture");
const compileData_1 = require("./adapters/compileData");
const comptoirDuLibreApi_1 = require("./adapters/comptoirDuLibreApi");
const dbApi_1 = require("./adapters/dbApi");
const getCnllPrestatairesSill_1 = require("./adapters/getCnllPrestatairesSill");
const getServiceProviders_1 = require("./adapters/getServiceProviders");
const getSoftwareLatestVersion_1 = require("./adapters/getSoftwareLatestVersion");
const getWikidataSoftware_1 = require("./adapters/getWikidataSoftware");
const getWikidataSoftwareOptions_1 = require("./adapters/getWikidataSoftwareOptions");
const getHalSoftware_1 = require("./adapters/hal/getHalSoftware");
const getHalSoftwareOptions_1 = require("./adapters/hal/getHalSoftwareOptions");
const userApi_1 = require("./adapters/userApi");
const usecases_1 = require("./usecases");
async function bootstrapCore(params) {
    const { gitDbApiParams, keycloakUserApiParams, githubPersonalAccessTokenForApiRateLimit, doPerPerformPeriodicalCompilation, doPerformCacheInitialization, externalSoftwareDataOrigin } = params;
    const { getSoftwareLatestVersion } = (0, getSoftwareLatestVersion_1.createGetSoftwareLatestVersion)({
        githubPersonalAccessTokenForApiRateLimit
    });
    const { getSoftwareExternalDataOptions, getSoftwareExternalData } = getSoftwareExternalDataFunctions(externalSoftwareDataOrigin);
    const { compileData } = (0, compileData_1.createCompileData)({
        getSoftwareExternalData,
        getCnllPrestatairesSill: getCnllPrestatairesSill_1.getCnllPrestatairesSill,
        comptoirDuLibreApi: comptoirDuLibreApi_1.comptoirDuLibreApi,
        getSoftwareLatestVersion,
        getServiceProviders: getServiceProviders_1.getServiceProviders
    });
    const { dbApi, initializeDbApiCache } = (0, dbApi_1.createGitDbApi)(gitDbApiParams);
    const { userApi, initializeUserApiCache } = keycloakUserApiParams === undefined
        ? {
            "userApi": (0, redux_clean_architecture_1.createObjectThatThrowsIfAccessed)({
                "debugMessage": "No Keycloak server"
            }),
            "initializeUserApiCache": async () => { }
        }
        : (0, userApi_1.createKeycloakUserApi)(keycloakUserApiParams);
    const context = {
        "paramsOfBootstrapCore": params,
        dbApi,
        userApi,
        compileData,
        comptoirDuLibreApi: comptoirDuLibreApi_1.comptoirDuLibreApi,
        getSoftwareExternalDataOptions,
        getSoftwareExternalData,
        getSoftwareLatestVersion
    };
    const { core, dispatch } = (0, redux_clean_architecture_1.createCore)({
        context,
        usecases: usecases_1.usecases
    });
    await dispatch(usecases_1.usecases.readWriteSillData.protectedThunks.initialize({
        doPerPerformPeriodicalCompilation
    }));
    if (doPerformCacheInitialization) {
        console.log("Performing cache initialization...");
        await Promise.all([initializeDbApiCache(), initializeUserApiCache()]);
    }
    return { core, context };
}
exports.bootstrapCore = bootstrapCore;
function getSoftwareExternalDataFunctions(externalSoftwareDataOrigin) {
    switch (externalSoftwareDataOrigin) {
        case "wikidata":
            return {
                "getSoftwareExternalDataOptions": getWikidataSoftwareOptions_1.getWikidataSoftwareOptions,
                "getSoftwareExternalData": getWikidataSoftware_1.getWikidataSoftware
            };
        case "HAL":
            return {
                "getSoftwareExternalDataOptions": getHalSoftwareOptions_1.getHalSoftwareOptions,
                "getSoftwareExternalData": getHalSoftware_1.getHalSoftware
            };
        default:
            const unreachableCase = externalSoftwareDataOrigin;
            throw new Error(`Unreachable case: ${unreachableCase}`);
    }
}
//# sourceMappingURL=bootstrap.js.map